;;;; -*- lisp -*-

(in-package :it.bese.yaclml)

(defun lookup-tal-variable (name enviroment)
  (getf enviroment name))

(defun read-tal-expression-from-string (expression enviroment-symbol)
  (let ((*readtable* (copy-readtable nil)))
    (declare (special *readtable*))
    ;; use $SYMBOL to access the value of the enviroment variable
    ;; SYMBOL
    (set-macro-character #\$ (lambda (stream char)
                               (declare (ignore char))
                               `(lookup-tal-variable ,(intern (string (read stream)) :keyword) ,enviroment-symbol))
                         nil *readtable*)
    (read-from-string expression)))

(defun normalize-lhtml (lhtml)
  "Convert the LHTML code into a more consistent form."
  (etypecase lhtml
    (keyword (list (list lhtml)))
    (string lhtml)
    (list (let ((tag-name (car lhtml))
                (body (cdr lhtml)))
            (cons (ensure-list tag-name) body)))))

(defvar *tal-attribute-handlers* '())

(defvar *tal-tag-handlers* '())

(defmacro def-attribute-handler (attribute (tag env) &body body)
  `(progn
     (push (cons (string ',attribute)
                 (lambda (,tag ,env) ,@body))
           *tal-attribute-handlers*)
     ',attribute))

(defmacro def-tag-handler (tag-name (tag env) &body body)
  `(progn
     (push (cons (string ',tag-name)
                 (lambda (,tag ,env) ,@body))
           *tal-tag-handlers*)
     ',tag-name))

(defun transform-lhtml-form (form enviroment-symbol)
  (setf form (normalize-lhtml form))
  (if (stringp form)
      `(<:as-is ,form)
      (destructuring-bind ((tag-name &rest attributes) &rest body) form
	;; first see if there's a handler for this tag
	(dolist* ((name . handler) *tal-tag-handlers*)
	  (when (string= name tag-name)
	    (return-from transform-lhtml-form
	      (funcall handler form enviroment-symbol))))
	;; nope, check the attributes
	(loop for (key value) on attributes by #'cddr
	      for handler = (assoc key *tal-attribute-handlers* :test #'string=)
	      when handler
	      do (return-from transform-lhtml-form
		   (funcall (cdr handler) form enviroment-symbol)))
	;; check and see if it's the "magic" TAL tag
	(when (eql :TAL tag-name)
	  ;; ignore the tag and just transform the body
	  (return-from transform-lhtml-form 
	    `(progn
	       ,@(transform-lhtml-tree body enviroment-symbol))))
	;; didn't find a handler for that tag or any of it's
	;; attributes, must be a "regular" yaclml tag.
	`(,(intern (string tag-name) :it.bese.yaclml.tags)
	  ,@attributes ,@(transform-lhtml-tree body enviroment-symbol)))))

(defun transform-lhtml-tree (tree enviroment-symbol)
  "Given a tree representing some LHTML code with TAL attributes
  returns the yaclml tag using code for generating the
  HTML. Destructivly modifies TREE."
  ;; we collect strings and forms it the collector OPS. When we're
  ;; done iterating (or recusring) over the tags we can string-fold
  ;; ops to get the longest posible string sequences.
  (mapcar (lambda (form)
	    (transform-lhtml-form form enviroment-symbol))
	  tree))

(def-attribute-handler :TAL\:CONTENT (tag enviroment-symbol)
  "Becomes a TAL:REPLACE."
  (let ((value (getf (cdar tag) :TAL\:CONTENT))
        (escape-html (getf (cdar tag) :TAL\:ESCAPE-HTML)))
    (remf (cdar tag) :TAL\:CONTENT)
    (remf (cdar tag) :TAL\:ESCAPE-HTML)
    (destructuring-bind ((tag-name &rest attributes) &rest body) tag
      (declare (ignore body))
      (transform-lhtml-form `((,tag-name ,@attributes)
			      ((:CONTENT-DUMMY :TAL\:REPLACE ,value :TAL\:ESCAPE-HTML ,escape-html) "DUMMY"))
			    enviroment-symbol))))

(def-attribute-handler :TAL\:REPLACE (tag enviroment-symbol)
  (let ((value (read-tal-expression-from-string (getf (cdar tag) :TAL\:REPLACE)
                                                enviroment-symbol))
        (escape (if-bind escape (getf (cdar tag) :TAL\:ESCAPE-HTML)
                  ;; if they supplied a vlaue then use it (either nil or t or whatever)
                  (read-tal-expression-from-string escape enviroment-symbol)
                  ;; no value supplied, default to T
                  t)))
    (remf (cdar tag) :TAL\:REPLACE)
    (remf (cdar tag) :TAL\:ESCAPE-HTML)
    (if escape
	`(<:as-html ,value)
        `(<:as-is ,value))))

(def-attribute-handler :TAL\:WHEN (tag enviroment-symbol)
  (let ((value (read-tal-expression-from-string (getf (cdar tag) :TAL\:WHEN)
                                                enviroment-symbol)))
    (remf (cdar tag) :TAL\:WHEN)
    `(when ,value
       ,(transform-lhtml-form tag enviroment-symbol))))

(def-attribute-handler :TAL\:DOLIST (tag enviroment-symbol)
  "On each iteration the enviroment is extended with the value in
  the value passed to DOLIST."
  (let ((value (read-tal-expression-from-string (getf (cdar tag) :TAL\:DOLIST) enviroment-symbol)))
    (remf (cdar tag) :TAL\:DOLIST)
    (with-unique-names (loop-item-sym)
      `(dolist (,loop-item-sym ,value)
	 (let ((,enviroment-symbol (nconc ,loop-item-sym ,enviroment-symbol)))
	   ,(transform-lhtml-form tag enviroment-symbol))))))
  
(defun compile-tal-string-to-lambda (string)
  (let ((output-stream (gensym))
	(enviroment (gensym)))
    `(lambda (,enviroment)
       (declare (ignorable ,enviroment))
       ,@(transform-lhtml-tree (parse-html string) enviroment))))

(defun compile-tal-string (string)
  (compile nil (compile-tal-string-to-lambda string)))

(defun compile-tal-file (pathname)
  (compile-tal-string (read-string-from-file pathname)))
